#include <sourcemod>
#include <sdktools>

#define DEBUG 0
#define PLUGIN_VERSION "1.1.2"

public Plugin:myinfo = 
{
	name = "L4D Rescue Tank",
	author = "ArmGadge",
	description = "Spawn tank from closet",
	version = PLUGIN_VERSION,
	url = "http://radiowave.moe.hm/"
};

new lastseed;
new ent_calling[MAXPLAYERS+1];
new closetstate[50];
new farest_surv_from_closet[50];

new Float:pos_teleport_tank[3];
new bool:spawnticket_tank;
new bool:reservewin;

new Handle:h_cv_probability = INVALID_HANDLE;
new Handle:h_cv_limit = INVALID_HANDLE;
new Handle:h_cv_announce = INVALID_HANDLE;
new Handle:h_ticketexpire = INVALID_HANDLE;

new Handle:h_eid_closet;	//adt_array クローゼットのEntity ID
new Handle:h_pos_closet;	//adt_array クローゼットの位置(vector)
new Handle:h_closet_group; //adt_array グループ化したクローゼットのEntity ID
new Handle:h_closet_arrindex; //adt_array h_eid_closet, h_pos_closetのインデックス
new Handle:closetid_to_group;	//adt_trie クローゼットのEntity IDとグループ番号
new Handle:h_eid_to_pos;	//adt_trie クローゼットのEntity IDと位置(vector)

new String:logpath[255];

public OnPluginStart()
{
	CreateConVar("l4d_rescuetank_version", PLUGIN_VERSION, "Version number of Rescue Tank plugin.",
		FCVAR_DONTRECORD|FCVAR_REPLICATED|FCVAR_NOTIFY);
	h_cv_probability = CreateConVar("l4d_rescuetank_probability", "20",
		"Percentage of spawn tank from closet (int range:0-100)[%]", 0, true, 0.0, true, 100.0);
	h_cv_limit = CreateConVar("l4d_rescuetank_survivor_limit", "0",
		"Spawn Tank from closet only when count of standing survivors higher than this value (int)",
		0, true, 0.0, false);
	h_cv_announce = CreateConVar("l4d_rescuetank_announce", "1", "Display message to hint-text when closet tank spawned.",
		0, true, 0.0, true, 1.0);
	
	AutoExecConfig(true, "l4d_rescuetank");
	
	HookEvent("survivor_call_for_help", Event_CallHelp);
	HookEvent("survivor_rescued", Event_Rescued);
	HookEvent("tank_spawn", Event_TankSpawn);
	HookEvent("round_start", Event_RoundStart);
	HookEvent("player_spawn", Event_PlayerSpawn);
	
	BuildPath(Path_SM, logpath, sizeof(logpath), "logs/l4d_rescuetank.log");	
	
	h_eid_closet = CreateArray(1);
	h_pos_closet = CreateArray(3);
	h_closet_group = CreateArray(50);	
	h_closet_arrindex = CreateArray(1);
	h_eid_to_pos = CreateTrie();
	closetid_to_group = CreateTrie();
}


public OnPluginEnd()
{
	CloseHandle(h_eid_closet);
	CloseHandle(h_pos_closet);
	CloseHandle(h_eid_to_pos);
	CloseHandle(h_closet_group);
	CloseHandle(h_closet_arrindex);
	CloseHandle(closetid_to_group);
}

public OnMapStart()
{
	new time_cur = GetTime();
	new timediff = time_cur-lastseed;
	if(timediff >= 10800)	//3hours
	{
		SetRandomSeed(time_cur);
		lastseed = time_cur;
		PrintToServer("[RescueTank] Reset random seed. (%d seconds past from last reset)", timediff);
	}
}

public OnMapEnd()
{
	if(h_ticketexpire != INVALID_HANDLE)
	{
		KillTimer(h_ticketexpire);
		h_ticketexpire = INVALID_HANDLE;
	}
}

public Event_RoundStart(Handle:event, const String:name[], bool:dontBroadcast)
{
	for (new i=1; i<=MaxClients; i++) ent_calling[i] = -1;
	for (new i=0; i<50; i++) closetstate[i] = 0;
	reservewin = false;
	DetectRescueCloset();
	//CancelReserve();
}

DetectRescueCloset()
{
	new ent;
	new Float:dist_vert;
	decl String:strbuf[32];
	decl Float:pos_closet_buf[3];
	
	ClearArray(h_eid_closet);
	ClearArray(h_pos_closet);
	ClearArray(h_closet_group);
	ClearArray(h_closet_arrindex);
	ClearTrie(closetid_to_group);
	ClearTrie(h_eid_to_pos);

	#if DEBUG
	LogToFile(logpath, "Analyzing Rescue Closet...");
	#endif

	while ((ent = FindEntityByClassname(ent, "info_survivor_rescue")) != -1)
	{
		GetEntPropVector(ent, Prop_Data, "m_vecAbsOrigin", pos_closet_buf);
		PushArrayCell(h_eid_closet, ent);
		PushArrayArray(h_pos_closet, pos_closet_buf);
		IntToString(ent, strbuf, sizeof(strbuf));
		SetTrieArray(h_eid_to_pos, strbuf, pos_closet_buf, 3);
		
	}

	#if DEBUG
	LogToFile(logpath, "Found %d info_survivor_rescue", GetArraySize(h_eid_closet));
	#endif


	for (new i=0; i<GetArraySize(h_eid_closet); i++)
		PushArrayCell(h_closet_arrindex, i);
	
	decl Float:dist, Float:buf1_origin[3], Float:buf2_origin[3];
	decl String:str_entid[8];new groupid;
	new index;
	//クローゼットのグループを分析
	while(GetArraySize(h_closet_arrindex) != 0)
	{
		#if DEBUG
		LogToFile(logpath, "Group ID %d ---", groupid);
		#endif
		//最初に一個入れる
		new arraybuf[150];
		for (new i=0; i<150; i++)
			arraybuf[i] = -1;
		index = GetArrayCell(h_closet_arrindex, 0);
		arraybuf[0] = GetArrayCell(h_eid_closet, index);
		#if DEBUG
		LogToFile(logpath, "Entity ID: %d", arraybuf[0]);
		#endif
		IntToString(arraybuf[0], str_entid, sizeof(str_entid));
		SetTrieValue(closetid_to_group, str_entid, groupid);
		GetArrayArray(h_pos_closet, index, buf1_origin);
		RemoveFromArray(h_closet_arrindex, 0);
		

		new i=0, count=1;
		/*
			上の一個と他のレスキューポイントの距離をみて、近ければグループに加える
			このときにarrindexのものを削除して、次も同じインデックスで見る
			近くなければ次のインデックスを見る
		*/
		while(i<GetArraySize(h_closet_arrindex))
		{
			index = GetArrayCell(h_closet_arrindex, i);
			GetArrayArray(h_pos_closet, index, buf2_origin);
			dist = GetVectorDistance(buf1_origin, buf2_origin);
			dist_vert = FloatAbs(buf1_origin[2] - buf2_origin[2]);
			//距離が近いかつ　z軸（上下）を比較　立体構造対策
			if (dist <= 300.0 && dist_vert <= 50.0)
			{
				ent = GetArrayCell(h_eid_closet, index);
				IntToString(ent, str_entid, sizeof(str_entid));
				//key:Entity ID of closet, value:group id
				SetTrieValue(closetid_to_group, str_entid, groupid);
				arraybuf[count] = ent;
				count++;
				RemoveFromArray(h_closet_arrindex, i);
				#if DEBUG
				LogToFile(logpath, "Entity ID: %d", ent);
				#endif
			}
			else i++;
		}
		PushArrayArray(h_closet_group, arraybuf);
		
		groupid++;


	}
	#if DEBUG
	LogToFile(logpath, "Analyzing Done.", groupid);
	#endif
}

public Event_CallHelp(Handle:event, const String:name[], bool:dontBroadcast)
{
	//救助待ちの人が呼ばれるたびに繰り返しトリガーがかかる
	new userid = GetEventInt(event, "userid");
	new client = GetClientOfUserId(userid); 
	new ent = GetEventInt(event, "subject");
	decl String:strbuf[32];
	new groupid;
	new count_waiting;
	// 以前呼ばれていたentidと違う、もしくは新規の場合
	if (ent != ent_calling[client])
	{
		//古いところにまだ人がいるか？
		if (ent_calling[client] != -1)
		{
			IntToString(ent_calling[client], strbuf, sizeof(strbuf));
			GetTrieValue(closetid_to_group, strbuf, groupid);
			#if DEBUG
			LogToFile(logpath, "Client %d: moved rescue point, old group was %d",
				client, groupid);
			#endif
			//当たりのクローゼットだった場合
			if (closetstate[groupid] == 1)
			{
				count_waiting = getCountOfCallingSurvivors(groupid);
				if (count_waiting == 0)
				{
					//誰もいなければ次の抽選は絶対当たりとする。
					reservewin = true;
					closetstate[groupid] = 0;
					#if DEBUG
					LogToFile(logpath, "Group %d have won the lottery, reserved next closet", groupid);
					#endif
				}
			}
			
		}
		//新しい方に人がいるかチェック
		IntToString(ent, strbuf, sizeof(strbuf));
		GetTrieValue(closetid_to_group, strbuf, groupid);
		if (getCountOfAliveSurvivors() <= GetConVarInt(h_cv_limit))
		{
			closetstate[groupid] = -1;
			#if DEBUG
			LogToFile(logpath, "Closet group %d have lost the lottery (Below the limit). ", groupid);
			#endif
		}		
		else if (closetstate[groupid] == 0)
		{
			//当たり予約がある場合は抽選せず当たり
			if (reservewin)
			{
				closetstate[groupid] = 1;
				reservewin = false;
				#if DEBUG
				LogToFile(logpath, "Closet group %d got reserved winning lottery.", groupid);
				#endif
			}
			else
			{
				//予約なければ抽選
				new random = GetRandomInt(1,100);
				if (random <= GetConVarInt(h_cv_probability))
				{
					closetstate[groupid] = 1;
					#if DEBUG
					LogToFile(logpath, "Closet group %d have won the lottery.", groupid);
					#endif
				}
				else
				{
					closetstate[groupid] = -1;
					#if DEBUG
					LogToFile(logpath, "Closet group %d have lost the lottery.", groupid);
					#endif
				}
			}
		}		
		ent_calling[client] = ent;
	}
	
	//クローゼットの座標から一番遠い生存者を登録しておく
	new Float:origin_closet[3], Float:origin_survivor[3], Float:dist;
	new Float:farestdist;
	new farestclient;
	
	IntToString(ent, strbuf, sizeof(strbuf));
	GetTrieArray(h_eid_to_pos, strbuf, origin_closet, 3);
	for (new i=1; i<=MaxClients; i++)
	{
		if(IsClientConnected(i) &&
			IsClientInGame(i) &&
			(GetClientTeam(i) == 2) &&
			IsPlayerAlive(i))
		{
			GetClientAbsOrigin(i, origin_survivor);
			dist = GetVectorDistance(origin_closet, origin_survivor);
			if (farestdist < dist)
			{
				farestdist = dist;
				farestclient = i;
			}
		}
	}
	farest_surv_from_closet[groupid] = farestclient;
}

public Event_PlayerSpawn(Handle:event, const String:name[], bool:dontBroadcast)
{
	new userid = GetEventInt(event, "userid");
	new client = GetClientOfUserId(userid);
	if (GetClientTeam(client) == 2)
		CreateTimer(2.0, SpawnDelay, client);
}
public Event_Rescued(Handle:event, const String:name[], bool:dontBroadcast)
{
	new uid_rescuer = GetEventInt(event, "rescuer");
	new uid_victim = GetEventInt(event, "victim");
	new rescuer = GetClientOfUserId(uid_rescuer);
	new victim = GetClientOfUserId(uid_victim);
	new Float:pos_teleport_surv[3];
	new groupid;
	decl String:strbuf[32];
	IntToString(ent_calling[victim], strbuf, sizeof(strbuf));
	GetTrieValue(closetid_to_group, strbuf, groupid);
	if (closetstate[groupid] == 1)
	{
		GetClientAbsOrigin(farest_surv_from_closet[groupid] , pos_teleport_surv);
		TeleportEntity(victim, pos_teleport_surv, NULL_VECTOR, NULL_VECTOR);
		#if DEBUG
		LogToFile(logpath, "Client %d belonged to tank closet group %d. teleported to client %d.",
		victim, groupid, farest_surv_from_closet[groupid]);
		#endif
		if (h_ticketexpire == INVALID_HANDLE)
		{
			IntToString(ent_calling[victim], strbuf, sizeof(strbuf));
			GetTrieArray(h_eid_to_pos, strbuf, pos_teleport_tank, 3);
			spawnticket_tank = true;
			h_ticketexpire = CreateTimer(1.0, TicketExpire, groupid);
			#if DEBUG
			LogToFile(logpath, "Sending Tank spawn command with ticket.");
			#endif
			ExecCheatCmd(rescuer, "z_spawn", "tank", "\0");
		}
		
	}
	else if (closetstate[groupid] == -1)
		closetstate[groupid] = 0;
	ent_calling[victim] = -1;
}

public Event_TankSpawn(Handle:event, const String:name[], bool:dontBroadcast)
{
	new playerid = GetEventInt(event, "userid");
	new client = GetClientOfUserId(playerid);
	if(spawnticket_tank)
	{
		TeleportEntity(client, pos_teleport_tank, NULL_VECTOR, NULL_VECTOR);
		spawnticket_tank = false;
		if (GetConVarBool(h_cv_announce))
		{
			CreateTimer(2.0, AnnounceClosetTank);
		}
		PrintToServer("[RescueTank] Tank has been rescued.");
		#if DEBUG
		LogToFile(logpath, "Confirmed spawn, teleported to closet.");
		#endif
	}
	//複数人救出があった場合は複数Tankになってしまうので、チケットは期限切れまで残しておく
}

ExecCheatCmd(client, String:command[], String:arg1[], String:arg2[])
{
	new flags = GetCommandFlags(command);
	SetCommandFlags(command, flags & ~FCVAR_CHEAT);
	if(client != 0) FakeClientCommand(client, "%s %s %s", command, arg1, arg2);
	else ServerCommand("%s %s %s", command, arg1, arg2);
	SetCommandFlags(command, flags);
}

public Action:TicketExpire(Handle:timer, any:groupid)
{
	spawnticket_tank = false;
	closetstate[groupid] = 0;
	h_ticketexpire = INVALID_HANDLE;
}

public Action:AnnounceClosetTank(Handle:timer)
{
	PrintHintTextToAll("///   Closet Tank!   ///");
}

getCountOfAliveSurvivors()
{
	new count;
	for (new i=1; i<=MaxClients; i++)
	{
		if (IsClientConnected(i) &&
			IsClientInGame(i) &&
			GetClientTeam(i) == 2 &&
			IsPlayerAlive(i))
		{
			count++;
		}
	}
	return count;
}

getCountOfCallingSurvivors(groupid)
{
	new ent_rescue[MAXPLAYERS+1];
	new count;
	GetArrayArray(h_closet_group, groupid, ent_rescue);
	for (new i=0; i<MAXPLAYERS+1; i++)
	{
		if (ent_rescue[i] == -1) break;
		//他に待ってる人を調べる
		new cid_call = GetEntPropEnt(ent_rescue[i], Prop_Send, "m_survivor");
		#if DEBUG
		LogToFile(logpath, "rescue point entity %d, m_survivor: %d",
			ent_rescue[i] ,cid_call);
		#endif
		if (cid_call > 0) count++;
	}
	#if DEBUG
	LogToFile(logpath, "getCountOfCallingSurvivors: Group %d have %d survivors.",
		groupid, count);
	#endif
	return count;
}

public Action:SpawnDelay(Handle:timer, any:client)
{
	new groupid;
	new count_waiting;
	decl String:strbuf[32];
	
	if (ent_calling[client] != -1)
	{
		#if DEBUG
		LogToFile(logpath, "Maybe client %d force respawned",
			client);
		#endif
		
		//元いたクローゼットに人が居るか
		IntToString(ent_calling[client], strbuf, sizeof(strbuf));
		GetTrieValue(closetid_to_group, strbuf, groupid);
		//当たりのクローゼットだった場合
		if (closetstate[groupid] == 1)
		{
			count_waiting = getCountOfCallingSurvivors(groupid);
			if (count_waiting == 0)
			{
				//誰もいなければ次の抽選は絶対当たりとする。
				reservewin = true;
				closetstate[groupid] = 0;
				
				#if DEBUG
				LogToFile(logpath, "Group %d have won the lottery, reserved next closet", groupid);
				#endif
			}
		}

		ent_calling[client] = -1;		
	}

}